#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/mm.h>

#include <linux/cooperative_internal.h>
#include <asm/cooperative_internal.h>
#include <asm/smp.h>
#include <asm/desc.h>
#include <asm/mmu_context.h>
#include <asm/debugreg.h>
#include <asm/i387.h>

CO_TRACE_STOP;


/*
 * The next asm code is the first Linux code that runs in the 
 * coLinux kernel context. It receives %ecx which contains the 
 * address of the passage page. The passage page code sets %ecx 
 * to this value in its context restore part.
 */  

asm(
	""
	".section .text\n"
	".globl co_start\n"
	"co_start:\n"
	"       call co_start_arch\n"
	".previous\n"
	"");

static int co_passage_page_holding_count = 0;

static void co_early_cpu_init(void)
{
	/*
	 * On the first switch to Linux we must set up a valid TR because 
	 * the passage page code assumes such one exists. This is basically
	 * copied code from cpu_init().
	 *
	 * P.S this is protected by CO_TRACE_STOP so that we don't
	 * have a monitor context switch.
	 */
        int cpu = smp_processor_id();
        struct tss_struct * t = &per_cpu(init_tss, cpu);
        struct thread_struct *thread = &current->thread;

	/*
	 * Initialize the per-CPU GDT with the boot GDT,
	 * and set up the GDT descriptor:
	 */
	memcpy(&per_cpu(cpu_gdt_table, cpu), cpu_gdt_table, GDT_SIZE);
	cpu_gdt_descr[cpu].size = GDT_SIZE - 1;
	cpu_gdt_descr[cpu].address = (unsigned long)&per_cpu(cpu_gdt_table, cpu);

        /*
         * Set up the per-thread TLS descriptor cache:
         */
        memcpy(thread->tls_array, &per_cpu(cpu_gdt_table, cpu), GDT_ENTRY_TLS_ENTRIES * 8);

        __asm__ __volatile__("lgdt %0" : : "m" (cpu_gdt_descr[cpu]));
        __asm__ __volatile__("lidt %0" : : "m" (idt_descr));

        /*
         * Delete NT
         */
        __asm__("pushfl ; andl $0xffffbfff,(%esp) ; popfl");

        /*
         * Set up and load the per-CPU TSS and LDT
         */
        atomic_inc(&init_mm.mm_count);
        current->active_mm = &init_mm;
        enter_lazy_tlb(&init_mm, current);

        load_esp0(t, thread);
        set_tss_desc(cpu,t);
        per_cpu(cpu_gdt_table, cpu)[GDT_ENTRY_TSS].b &= 0xfffffdff;

        load_TR_desc();

        load_LDT(&init_mm.context);

        /* Set up doublefault TSS pointer in the GDT */
        __set_tss_desc(cpu, GDT_ENTRY_DOUBLEFAULT_TSS, &doublefault_tss);
        per_cpu(cpu_gdt_table, cpu)[GDT_ENTRY_DOUBLEFAULT_TSS].b &= 0xfffffdff;

        /* Clear %fs and %gs. */
        asm volatile ("xorl %eax, %eax; movl %eax, %fs; movl %eax, %gs");

	__asm__ __volatile__("movl %%cr4, %0" : "=r" (mmu_cr4_features));
}

asm(
	""
	".section .text\n"
	".globl co_arch_start_kernel\n"
	"co_arch_start_kernel:\n"
	"       call co_startup_entry\n"
	".previous\n"
	"");

void co_start_arch(void)
{
	co_early_cpu_init();
	co_start_kernel();
}

extern void ctrl_alt_del(void);

void co_handle_jiffies(long count)
{
	unsigned long flags;
	struct pt_regs regs;

	if (count > HZ) {
		xtime.tv_sec += count / HZ;
		count -= ((count / HZ) * HZ);
	}
		
	while (count > 0) {
		local_irq_save(flags);
		regs.orig_eax = TIMER_IRQ;
		do_IRQ(&regs);
		local_irq_restore(flags);

		count--;
	}
}

void co_handle_incoming_message(co_message_node_t *node_message)
{
	unsigned long flags;
	struct pt_regs regs;
	co_linux_message_t *message;

	message = (co_linux_message_t *)&node_message->msg.data;

	switch (message->device) {
	case CO_DEVICE_POWER: {
		co_linux_message_power_t *type = (co_linux_message_power_t *)message->data;
		switch (type->type) {
		case CO_LINUX_MESSAGE_POWER_ALT_CTRL_DEL: {
			ctrl_alt_del();
			break;
		}
		}
		co_free_message(node_message);
		break;
	}

	case CO_DEVICE_KEYBOARD: {
		co_queue_incoming_message(node_message);

		local_irq_save(flags);
		regs.orig_eax = KEYBOARD_IRQ;
		do_IRQ(&regs);
		local_irq_restore(flags);
		break;
	}
	
	case CO_DEVICE_NETWORK: {
		co_queue_incoming_message(node_message);

		local_irq_save(flags);
		regs.orig_eax = NETWORK_IRQ;
		do_IRQ(&regs);
		local_irq_restore(flags);
		break;
	}

	case CO_DEVICE_SERIAL: {
		co_queue_incoming_message(node_message);

		local_irq_save(flags);
		cocd_interrupt();
		local_irq_restore(flags);
		break;
	}

	default:
		co_free_message(node_message);
		break;
	}
}

void co_switch_wrapper_protected(void)
{
	kernel_fpu_begin();

	/*
	 * We don't trust the passage page code to safely restore %gs and %fs. 
	 *
	 * This wrapper ensures that if %fs or %gs are invalid, the processes
	 * exits with a segmentation fault rather than bringing down the 
	 * machine.
	 **/
	unsigned long fs = 0;
	unsigned long gs = 0;

        asm volatile("movl %%fs,%0": "=m" (fs));
        asm volatile("movl %%gs,%0": "=m" (gs));

	/*
	 * Nullify the registers so the passage page code restores to 
	 * null segment values on return.
	 */
        asm volatile("movl %0, %%fs;  movl %0, %%gs" : : "r" (0));

	/* And switch... */
	co_switch();

	/*
	 * Safely restore the registers.
	 */
	loadsegment(fs, fs);
	loadsegment(gs, gs);

	kernel_fpu_end();
}

void co_switch_wrapper(void)
{
	/* taken from irq.c: debugging check for stack overflow */
	long esp;
	
	__asm__ __volatile__("andl %%esp,%0" : "=r" (esp) : "0" (THREAD_SIZE - 1));
	if (unlikely(esp < (sizeof(struct thread_info) + STACK_WARN))) {
		printk("co_switch_wrapper: stack overflow: %ld\n", esp - sizeof(struct thread_info));
		co_terminate(CO_TERMINATE_STACK_OVERFLOW);
	}

	co_switch_wrapper_protected();
}

void co_passage_page_acquire(unsigned long *flags)
{
	local_irq_save(*flags);
	co_passage_page_holding_count++;
}

int co_passage_page_held(void)
{
	return co_passage_page_holding_count;
}

void co_passage_page_release(unsigned long flags)
{
	co_passage_page_holding_count--;
	local_irq_restore(flags);
}

void co_debug(const char *fmt, ...)
{
}

#define MAX_TRACE_POINTS 1024

typedef struct {	
	unsigned char *code;
	unsigned char original_byte;
	int off;
} co_tracepoint_t;

co_tracepoint_t tracepoints[MAX_TRACE_POINTS];
static int active_tracepoints = 0;

void co_kernel_breakpoint(struct pt_regs * regs)
{
	int i = 0;
	unsigned char *code = (unsigned char *)regs->eip;
	if (!code)
		return;

	for (i=0; i < active_tracepoints; i++) {
		if (tracepoints[i].code == code - 1) {
			co_debug("TRACEPOINT: %x\n", code - 1);
			break;
		}
	}

	if (i == active_tracepoints) {
		/* Bad, we don't know this tracepoint */
		co_terminate(CO_TERMINATE_INVALID_OPERATION);
		return;
	}

	*tracepoints[i].code = tracepoints[i].original_byte;
	regs->eflags |= (1 << 8); /* Enable TF */
	regs->eip = (unsigned long)(code - 1);
	tracepoints[i].off = 1;
}

void co_kernel_set_breakpoints(void)
{
	int i;

	for (i=0; i < active_tracepoints; i++)
		if (tracepoints[i].code  &&  tracepoints[i].off) {
			*tracepoints[i].code = 0xcc;
			tracepoints[i].off = 0;
		}
}

int co_kernel_debug(struct pt_regs *regs, long error_code, unsigned int condition)
{
	/* if not a single step trap */
	if (!(condition & DR_STEP))
		return 0;

	/* if userspace */
	if (regs->xcs & 3)
		return 0;

	regs->eflags &= ~(1 << 8); /* Disable TF */

	co_kernel_set_breakpoints();
	
	return 1;
}

void co_kernel_tracepoint_add(unsigned char *code)
{
	if (active_tracepoints >= MAX_TRACE_POINTS)
		return;

	tracepoints[active_tracepoints].code = code;
	tracepoints[active_tracepoints].original_byte = *code;
	tracepoints[active_tracepoints].off = 0;
	active_tracepoints++;
	*code = 0xcc;
}

co_arch_info_t co_arch_info = {
	.kernel_cs = __KERNEL_CS,
	.kernel_ds = __KERNEL_DS,
};

CO_TRACE_CONTINUE;
