/*
 *  Cooperative mode timer.
 *
 *  Dan Aloni <da-x@colinux.org>, 2003-2004 (C).
 */

#include <linux/init.h>
#include <linux/errno.h>

#include <asm/timer.h>
#include <asm/cooperative.h>
#include <asm/div64.h>
#include <asm/param.h>

#include <linux/cooperative.h>
#include <linux/cooperative_internal.h>

static unsigned long long first_time;
static unsigned long frequencey;
static unsigned long long last_mark, last_mark_quotient;
static unsigned long long last_time;

static unsigned long long query_host_highprec_time(void)
{
	unsigned long flags;
	unsigned long long this_time;
	unsigned long long diff;

	co_passage_page_assert_valid();	
	
	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_GET_HIGH_PREC_TIME;
	co_switch_wrapper();

	this_time = *(unsigned long long *)(&co_passage_page->params[0]);
	frequencey = *(unsigned long *)(&co_passage_page->params[2]);
	diff = ((long long)this_time - (long long)last_time);
	
	/*
	 * There shouldn't be any particularly large difference between
	 * the current and last host timestamps. For sanity, reset the 
         * global reference variables if we encounter any difference 
	 * larger than one second.
	 */

	if (diff < 0 || diff > frequencey) {
		first_time = this_time;
		last_mark_quotient = last_mark = 0;
	}

	last_time = this_time;
	co_passage_page_release(flags);

	return this_time;
}

static unsigned long long monotonic_clock_cooperative(void)
{
	return 0;
}

static long get_offset_cooperative(void)
{
	unsigned long flags;

	local_irq_save(flags);
	
	unsigned long long this_time = query_host_highprec_time() - first_time;
	unsigned long reminder = 0, result;
	long long diff, lldiff;
	long signed_result;

	diff = ((long long)this_time - (long long)(last_mark));
	if (diff < 0)
		lldiff = -diff;
	else
		lldiff = diff;
	
	lldiff *= 1000000;
	result = div_ll_X_l_rem(lldiff, frequencey, &reminder);

	signed_result = result;
	if (diff < 0)
		signed_result = -signed_result;

	local_irq_restore(flags);
	return signed_result;
}

static void mark_offset_cooperative(void)
{
	unsigned long flags;
	local_irq_save(flags);

	last_mark += frequencey / HZ;
	last_mark_quotient += frequencey % HZ;
	if (frequencey > HZ) {
		last_mark += 1;
		last_mark_quotient -= HZ;
	}

	local_irq_restore(flags);
}

static void delay_cooperative(unsigned long loops)
{
	/*
	 * A bogos delay loop for creating BogoMIPS...
	 */

	loops = loops / 10000;
	while (loops) {
		query_host_highprec_time();
		loops -= 1;
	}
}

static int __init init_cooperative_timer(char* override)
{
	first_time = query_host_highprec_time();

	/* Always pick this timer */
	return 0;
}

/************************************************************/

/* tsc timer_opts struct */
struct timer_opts timer_cooperative = {
	.name = 		"cooperative",
	.mark_offset =		mark_offset_cooperative,
	.get_offset =		get_offset_cooperative,
	.monotonic_clock =	monotonic_clock_cooperative,
	.delay = 		delay_cooperative,
};

struct init_timer_opts __initdata timer_cooperative_init = {
	.init =	init_cooperative_timer,
	.opts = &timer_cooperative,
};
