/*
 *  Cooperative Linux virtual keyboard controller driver
 *
 *  Copyright (c) 1999-2002 Dan Aloni <da-x@colinux.org)
 *    Based on 98kbd-io.c written by Osamu Tomita>
 */

/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/serio.h>
#include <linux/sched.h>
#include <linux/cooperative_internal.h>

#include <asm/io.h>

MODULE_AUTHOR("Dan Aloni <da-x@colinux.org>");
MODULE_DESCRIPTION("Cooperative Linux virtual keyboard controller driver");
MODULE_LICENSE("GPL");

/*
 * Names.
 */

#define COKBD_PHYS_DESC        "cokbd"

static struct serio cokbd_port;

static irqreturn_t cokbdio_interrupt(int irq, void *dev_id, struct pt_regs *regs);

/*
 * cokbd_flush() flushes all data that may be in the keyboard buffers
 */

static int cokbd_flush(void)
{
#if (0)	
	co_linux_message_t *message;

	while (co_get_message(&message, CO_DEVICE_KEYBOARD)) {
		co_free_message(message);
	}
#endif
	return 0;
}

/*
 * cokbd_write() sends a byte out through the keyboard interface.
 */

#define ATKBD_CMD_GETID		0x02f2

static void cokbd_receive(struct serio *port, unsigned char c)
{
	struct pt_regs regs= {0, };

	serio_interrupt(port, c, 0, &regs);
}

static int cokbd_write(struct serio *port, unsigned char c)
{
	return 0;
}

/*
 * cokbd_open() is called when a port is open by the higher layer.
 * It allocates the interrupt and enables in in the chip.
 */

static int cokbd_open(struct serio *port)
{
	cokbd_flush();

	if (request_irq(KEYBOARD_IRQ, cokbdio_interrupt, 0, "cokbd", NULL)) {
		printk(KERN_ERR "cobkd.c: Can't get irq %d for %s, unregistering the port.\n", KEYBOARD_IRQ, "KBD");
		serio_unregister_port(port);
		return -1;
	}

	return 0;
}

static void cokbd_close(struct serio *port)
{
	printk(KERN_INFO "cokbd closed\n");

	free_irq(KEYBOARD_IRQ, NULL);

	cokbd_flush();
}

/*
 * Structures for registering the devices in the serio.c module.
 */

static struct serio cokbd_port =
{
	.type =		SERIO_8042_XL,
	.write =	cokbd_write,
	.open =		cokbd_open,
	.close =	cokbd_close,
	.name =		"cokbd port",
	.phys =		COKBD_PHYS_DESC,
};

/*
 * cokbdio_interrupt() is the most important function in this driver -
 * it handles the interrupts from keyboard, and sends incoming bytes
 * to the upper layers.
 */

static irqreturn_t cokbdio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	co_message_node_t *node_message;
	while (co_get_message(&node_message, CO_DEVICE_KEYBOARD)) {
		co_linux_message_t *message = (co_linux_message_t *)&node_message->msg.data;
		co_scan_code_t *sc = (co_scan_code_t *)message->data;
		unsigned long scancode = sc->code;
		
		if (!sc->down)
			scancode |= 0x80;
		
		cokbd_receive(&cokbd_port, scancode);
		
		co_free_message(node_message);
	}
	
	return IRQ_HANDLED;
}

int __init cokbdio_init(void)
{
	serio_register_port(&cokbd_port);

	printk(KERN_INFO "serio: cokbd at irq %d\n", KEYBOARD_IRQ);

	return 0;
}

void __exit cokbdio_exit(void)
{
	serio_unregister_port(&cokbd_port);
}

module_init(cokbdio_init);
module_exit(cokbdio_exit);
