/*
 *  linux/kernel/cooperative.c
 *
 *  Cooperative mode (coLinux) support routines.
 *
 *  Dan Aloni <da-x@colinux.org>, 2003-2004 (C).
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/cooperative_internal.h>

CO_TRACE_STOP;

void start_kernel(void);
extern char _kernel_start, _end;


unsigned long co_core_end = 0;
unsigned long co_memory_size = 0;
void *co_initrd = NULL;
unsigned long co_initrd_size = 0;
char co_boot_parameters[CO_BOOTPARAM_STRING_LENGTH];


typedef struct {
	struct list_head list;
	int num_messages;
} co_message_queue_t;

int co_messages_active = 0;
co_message_queue_t co_outgoing_messages;
co_message_queue_t co_incoming_messages;
co_message_queue_t *co_incoming_queued_messages;

void co_start_kernel(void)
{
        co_core_end = co_passage_page->params[0];
        co_memory_size = co_passage_page->params[1];
        co_initrd = (void *)co_passage_page->params[2];
        co_initrd_size = co_passage_page->params[3];

        memcpy(co_boot_parameters, &co_passage_page->params[10], 
               sizeof(co_boot_parameters));

	co_arch_start_kernel();

	/* should never be reached */
	co_terminate(CO_TERMINATE_END);
}

co_message_t *co_send_message_save(unsigned long *flags)
{
	co_passage_page_assert_valid();	
	co_passage_page_acquire(flags);
	co_passage_page->operation = CO_OPERATION_MESSAGE_TO_MONITOR;
	return ((co_message_t *)CO_VPTR_IO_AREA_START);
}

void co_send_message_restore(unsigned long flags)
{
	co_switch_wrapper();
	co_passage_page_release(flags);
}

void co_send_message_s(co_message_t *message, const char *data)
{
	if ((sizeof(co_message_t) + message->size) > CO_VPTR_IO_AREA_SIZE)
		return;
	
	if (co_passage_page_held())
		return;
	
	unsigned long flags;
	co_message_t *buffer = ((co_message_t *)CO_VPTR_IO_AREA_START);
	
	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_MESSAGE_TO_MONITOR;
	*buffer = *message;
	memcpy(buffer->data, data, message->size);
	co_switch_wrapper();
	co_passage_page_release(flags);
}

void co_send_message(co_module_t from, 
		     co_module_t to,
		     co_priority_t priority,
		     co_message_type_t type,
		     unsigned long size,
		     const char *data)
{
	co_message_t params;

	params.from = from;
	params.to = to;
	params.priority = priority;
	params.type = type;
	params.size = size;

	co_send_message_s(&params, data);
}

static void co_message_add_to_incoming(co_message_t *message, unsigned long size)
{
	co_message_node_t *message_copy;

	message_copy = kmalloc(size + sizeof(co_message_node_t) - sizeof(co_message_t), 
			       GFP_ATOMIC);
	if (!message_copy)
		return;
	
	memcpy(&message_copy->msg, message, size);
	list_add_tail(&message_copy->node, &co_incoming_messages.list);
}

void co_callback(unsigned long flags)
{
	if (co_passage_page->operation != CO_OPERATION_MESSAGE_FROM_MONITOR) {
		co_passage_page_release(flags);
		return;
	}

	long io_size = co_passage_page->params[0];
	unsigned long new_jiffies = co_passage_page->params[1];

	if (co_messages_active  &&  io_size > 0) {
		unsigned char *io_buffer = (char *)CO_VPTR_IO_AREA_START;
		unsigned char *io_buffer_end = io_buffer + io_size;
		if (!(io_size > CO_VPTR_IO_AREA_SIZE)) {
			while (io_buffer < io_buffer_end) {
				co_message_t *message = (co_message_t *)io_buffer;
				unsigned long size = message->size + sizeof(*message);
				
				co_message_add_to_incoming(message, size);
				io_buffer += size;
			}
		}
	}

	co_passage_page_release(flags);

	co_handle_jiffies(new_jiffies);
	co_handle_incoming_messages();
}

void co_handle_incoming_messages(void)
{
	if (!co_messages_active)
		return;

	if (list_empty(&co_incoming_messages.list))
		return;

	for (;;) {
		unsigned long flags;
		co_message_node_t *message = NULL;

		/*
		 * Pop a message from the incoming queue.
		 */
		local_irq_save(flags);
		if (!list_empty(&co_incoming_messages.list)) {
			message = list_entry(co_incoming_messages.list.next, 
					     co_message_node_t, node);
			list_del(&message->node);
		}
		local_irq_restore(flags);

		if (!message)
			break;
		
		/*
		 * Let the interrupt routine of the arch dependant code
		 * handle the message, and be responsible to free it.
		 */
		co_handle_incoming_message(message);
	}
}

void co_idle_processor(void)
{
	unsigned long flags;

	co_passage_page_assert_valid();	

	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_IDLE;
	co_switch_wrapper();
	co_callback(flags);
}

void co_printk(const char *line)
{
	co_send_message(CO_MODULE_LINUX, 
			CO_MODULE_PRINTK,
			CO_PRIORITY_DISCARDABLE,
			CO_MESSAGE_TYPE_STRING,
			strlen(line)+1,
			line);
}

void co_debug_line(char *line)
{
}

void co_terminate(co_termination_reason_t reason)
{
	unsigned long flags;

	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_TERMINATE;
	co_passage_page->params[0] = reason;
	co_switch_wrapper();
	/* This doesn't really return. This code shouldn't be running. */
	co_passage_page_release(flags);
}

unsigned long co_get_host_time(void)
{
	unsigned long flags;
	unsigned long time;

	co_passage_page_assert_valid();

	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_GET_TIME;
	co_switch_wrapper();
	time = co_passage_page->params[0];
	co_passage_page_release(flags);

	return time;
}

void co_queue_incoming_message(co_message_node_t *node_message)
{
	if (!co_messages_active)
		return;
	
	co_linux_message_t *message = (co_linux_message_t *)&node_message->msg.data;
	if (message->device < 0 || (message->device >= CO_DEVICES_TOTAL))
		return;
	
	co_message_queue_t *queue;
	queue = &co_incoming_queued_messages[message->device];

	/* Add to the queue */
	unsigned long flags;
	local_irq_save(flags);
	list_add(&node_message->node, &queue->list);
	queue->num_messages++;
	local_irq_restore(flags);
}

int co_get_message(co_message_node_t **message, co_device_t device)
{
	co_message_queue_t *queue;
	co_message_node_t *node;
	unsigned long flags;

	if (!co_messages_active)
		return 0;

	local_irq_save(flags);
	queue = &co_incoming_queued_messages[device];
	if (list_empty(&queue->list)) {
		local_irq_restore(flags);
		return 0;
	}

	node = list_entry(queue->list.prev, co_message_node_t, node);
	list_del(&node->node);
	queue->num_messages--;
	local_irq_restore(flags);

	*message = node;
	return 1;
}

void co_free_message(co_message_node_t *message)
{
	kfree(message);
}

co_info_t co_info = {
	.api_version = CO_LINUX_API_VERSION,
	.compiler_major = __GNUC__,
	.compiler_minor = 3,
};

static int __init initcall_message_queues(void)
{
	int queue_index;

	INIT_LIST_HEAD(&co_outgoing_messages.list);
	INIT_LIST_HEAD(&co_incoming_messages.list);

	co_incoming_queued_messages = kmalloc(sizeof(co_message_queue_t) * CO_DEVICES_TOTAL, 
					      GFP_KERNEL);
	if (!co_incoming_queued_messages)
		panic("unable to allocate message queues\n");

	for (queue_index=0; queue_index < CO_DEVICES_TOTAL; queue_index++) {
		co_message_queue_t *queue = &co_incoming_queued_messages[queue_index];
		queue->num_messages = 0;
		INIT_LIST_HEAD(&queue->list);
	}

	co_messages_active = 1;

	return 0;
}


void co_free_pages(unsigned long vaddr, int order)
{
	unsigned long flags;
	
	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_FREE_PAGES;
	co_passage_page->params[0] = vaddr;
	co_passage_page->params[1] = order;
	co_switch_wrapper();
	co_passage_page_release(flags);
}

int co_alloc_pages(unsigned long vaddr, int order)
{
	unsigned long flags;
	long result;

	co_passage_page_acquire(&flags);
	co_passage_page->operation = CO_OPERATION_ALLOC_PAGES;
	co_passage_page->params[0] = vaddr;
	co_passage_page->params[1] = order;
	co_switch_wrapper();
	result = (long)co_passage_page->params[4];
	co_passage_page_release(flags);

	if (result < 0)
		return -ENOMEM;

	return 0;
}

__initcall(initcall_message_queues);

EXPORT_SYMBOL(co_terminate);

CO_TRACE_CONTINUE;
